@@ -40,6 +40,8 @@ enum event_type {
 	EVT_CLIENT_CONNECTED,
 	EVT_CLIENT_DISCONNECTED,
 	EVT_DETACHED,
+	EVT_OUTPUT_ADDED,
+	EVT_OUTPUT_REMOVED,
 	EVT_UNKNOWN,
 };
 #define EVT_LIST_LEN EVT_UNKNOWN
