@@ -20,6 +20,9 @@
 #include <libdrm/drm_fourcc.h>
 #include <stdbool.h>
 
+struct pixman_region16;
+
 enum wl_shm_format fourcc_to_wl_shm(uint32_t in);
 uint32_t fourcc_from_wl_shm(enum wl_shm_format in);
 int pixel_size_from_fourcc(uint32_t fourcc);
+uint32_t calculate_region_area(struct pixman_region16* region);
