@@ -15,6 +15,8 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
+#include <errno.h>
+#include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -26,25 +28,48 @@
 
 #include "data-control.h"
 
+static const char custom_mime_type_data[] = "wayvnc";
+
 struct receive_context {
-	struct data_control* data_control;
-	struct zwlr_data_control_offer_v1* offer;
+	struct nvnc* server;
+	struct aml_handler* handler;
+	LIST_ENTRY(receive_context) link;
 	int fd;
 	FILE* mem_fp;
 	size_t mem_size;
 	char* mem_data;
 };
 
-static void destroy_receive_context(void* raw_ctx)
+struct send_context {
+	struct aml_handler* handler;
+	LIST_ENTRY(send_context) link;
+	int fd;
+	char* data;
+	size_t length;
+	size_t index;
+};
+
+static void destroy_receive_context(struct receive_context* ctx)
 {
-	struct receive_context* ctx = raw_ctx;
-	int fd = ctx->fd;
+	aml_stop(aml_get_default(), ctx->handler);
+	aml_unref(ctx->handler);
 
 	if (ctx->mem_fp)
 		fclose(ctx->mem_fp);
 	free(ctx->mem_data);
-	zwlr_data_control_offer_v1_destroy(ctx->offer);
-	close(fd);
+	close(ctx->fd);
+	LIST_REMOVE(ctx, link);
+	free(ctx);
+}
+
+static void destroy_send_context(struct send_context* ctx)
+{
+	aml_stop(aml_get_default(), ctx->handler);
+	aml_unref(ctx->handler);
+
+	close(ctx->fd);
+	free(ctx->data);
+	LIST_REMOVE(ctx, link);
 	free(ctx);
 }
 
@@ -57,7 +82,12 @@ static void on_receive(void* handler)
 	char buf[4096];
 
 	ssize_t ret = read(fd, &buf, sizeof(buf));
-	if (ret > 0) {
+	if (ret == -1) {
+		if (errno == EAGAIN || errno == EWOULDBLOCK)
+			return;
+		nvnc_log(NVNC_LOG_ERROR, "Clipboard read failed: %m");
+		destroy_receive_context(ctx);
+	} else if (ret > 0) {
 		fwrite(&buf, 1, ret, ctx->mem_fp);
 		return;
 	}
@@ -66,10 +96,37 @@ static void on_receive(void* handler)
 	ctx->mem_fp = NULL;
 
 	if (ctx->mem_size)
-		nvnc_send_cut_text(ctx->data_control->server, ctx->mem_data,
-				ctx->mem_size);
+		nvnc_send_cut_text(ctx->server, ctx->mem_data, ctx->mem_size);
 
-	aml_stop(aml_get_default(), handler);
+	destroy_receive_context(ctx);
+}
+
+static void on_send(void* handler)
+{
+	struct send_context* ctx = aml_get_userdata(handler);
+	int fd = aml_get_fd(handler);
+	assert(ctx->fd == fd);
+
+	int ret;
+	ret = write(fd, ctx->data + ctx->index, ctx->length - ctx->index);
+	if (ret == -1) {
+		if (errno == EAGAIN || errno == EWOULDBLOCK)
+			return;
+		nvnc_log(NVNC_LOG_ERROR, "Clipboard write failed/incomplete: %m");
+		destroy_send_context(ctx);
+	} else if (ret == (int)(ctx->length - ctx->index)) {
+		destroy_send_context(ctx);
+	} else {
+		ctx->index += ret;
+	}
+}
+
+static int dont_block(int fd)
+{
+	int ret = fcntl(fd, F_GETFL);
+	if (ret == -1)
+		return -1;
+	return fcntl(fd, F_SETFL, ret | O_NONBLOCK);
 }
 
 static void receive_data(void* data,
@@ -83,39 +140,51 @@ static void receive_data(void* data,
 		return;
 	}
 
+	if (dont_block(pipe_fd[0]) == -1) {
+		nvnc_log(NVNC_LOG_ERROR, "Failed to set O_NONBLOCK on clipbooard receive fd");
+		close(pipe_fd[0]);
+		close(pipe_fd[1]);
+		return;
+	}
+
 	struct receive_context* ctx = calloc(1, sizeof(*ctx));
 	if (!ctx) {
-		nvnc_log(NVNC_LOG_ERROR, "OOM");
+		nvnc_log(NVNC_LOG_ERROR, "OOM: %m");
 		close(pipe_fd[0]);
 		close(pipe_fd[1]);
 		return;
 	}
 
 	zwlr_data_control_offer_v1_receive(offer, self->mime_type, pipe_fd[1]);
-	wl_display_flush(self->wl_display);
 	close(pipe_fd[1]);
 
 	ctx->fd = pipe_fd[0];
-	ctx->data_control = self;
-	ctx->offer = offer;
+	ctx->server = self->server;
 	ctx->mem_fp = open_memstream(&ctx->mem_data, &ctx->mem_size);
 	if (!ctx->mem_fp) {
-		close(ctx->fd);
-		free(ctx);
 		nvnc_log(NVNC_LOG_ERROR, "open_memstream() failed: %m");
-		return;
+		goto open_memstream_failure;
 	}
 
-	struct aml_handler* handler = aml_handler_new(ctx->fd, on_receive,
-			ctx, destroy_receive_context);
-	if (!handler) {
-		close(ctx->fd);
-		free(ctx);
-		return;
+	ctx->handler = aml_handler_new(ctx->fd, on_receive, ctx, NULL);
+	if (!ctx->handler) {
+		goto handler_failure;
 	}
 
-	aml_start(aml_get_default(), handler);
-	aml_unref(handler);
+	if (aml_start(aml_get_default(), ctx->handler) < 0) {
+		goto poll_start_failure;
+	}
+
+	LIST_INSERT_HEAD(&self->receive_contexts, ctx, link);
+	return;
+
+poll_start_failure:
+	aml_unref(ctx->handler);
+handler_failure:
+	fclose(ctx->mem_fp);
+open_memstream_failure:
+	free(ctx);
+	close(pipe_fd[0]);
 }
 
 static void data_control_offer(void* data,
@@ -124,13 +193,16 @@ static void data_control_offer(void* data,
 {
 	struct data_control* self = data;
 
-	if (self->offer)
-		return;
-	if (strcmp(mime_type, self->mime_type) != 0) {
+	if (strcmp(mime_type, self->custom_mime_type_name) == 0) {
+		self->is_own_offer = true;
 		return;
 	}
 
-	self->offer = zwlr_data_control_offer_v1;
+	if (self->offer)
+		return;
+
+	if (strcmp(mime_type, self->mime_type) == 0)
+		self->offer = zwlr_data_control_offer_v1;
 }
 
 struct zwlr_data_control_offer_v1_listener data_control_offer_listener = {
@@ -152,10 +224,22 @@ static void data_control_device_selection(void* data,
 	struct zwlr_data_control_offer_v1* id)
 {
 	struct data_control* self = data;
-	if (id && self->offer == id) {
-		receive_data(data, id);
-		self->offer = NULL;
+
+	if (!id) {
+		if (self->offer) {
+			zwlr_data_control_offer_v1_destroy(self->offer);
+			self->offer = NULL;
+			self->is_own_offer = false;
+		}
+		return;
 	}
+
+	if (id == self->offer && !self->is_own_offer)
+		receive_data(data, id);
+
+	zwlr_data_control_offer_v1_destroy(id);
+	self->offer = NULL;
+	self->is_own_offer = false;
 }
 
 static void data_control_device_finished(void* data,
@@ -169,11 +253,22 @@ static void data_control_device_primary_selection(void* data,
 	struct zwlr_data_control_offer_v1* id)
 {
 	struct data_control* self = data;
-	if (id && self->offer == id) {
-		receive_data(data, id);
-		self->offer = NULL;
+
+	if (!id) {
+		if (self->offer) {
+			zwlr_data_control_offer_v1_destroy(self->offer);
+			self->offer = NULL;
+			self->is_own_offer = false;
+		}
 		return;
 	}
+
+	if (id == self->offer && !self->is_own_offer)
+		receive_data(data, id);
+
+	zwlr_data_control_offer_v1_destroy(id);
+	self->offer = NULL;
+	self->is_own_offer = false;
 }
 
 static struct zwlr_data_control_device_v1_listener data_control_device_listener = {
@@ -190,18 +285,78 @@ data_control_source_send(void* data,
 	int32_t fd)
 {
 	struct data_control* self = data;
-	char* d = self->cb_data;
+	const char* d = self->cb_data;
 	size_t len = self->cb_len;
 	int ret;
 
 	assert(d);
+	assert(len);
+
+	if (strcmp(mime_type, self->custom_mime_type_name) == 0) {
+		d = custom_mime_type_data;
+		len = strlen(custom_mime_type_data);
+	}
+
+	if (dont_block(fd) == -1) {
+		nvnc_log(NVNC_LOG_ERROR, "Failed to set O_NONBLOCK on clipbooard send fd");
+		close(fd);
+		return;
+	}
 
 	ret = write(fd, d, len);
+	if (ret == -1) {
+		if (errno == EAGAIN || errno == EWOULDBLOCK) {
+			ret = 0;
+		} else {
+			nvnc_log(NVNC_LOG_ERROR, "Clipboard write failed: %m");
+			close(fd);
+			return;
+		}
+	} else if (ret == (int)len) {
+		close(fd);
+		return;
+	}
 
-	if (ret < (int)len)
-		nvnc_log(NVNC_LOG_ERROR, "write from clipboard incomplete");
+	/* we did a partial write, so continue sending data asynchronously */
 
+	struct send_context* ctx = calloc(1, sizeof(*ctx));
+	if (!ctx) {
+		nvnc_log(NVNC_LOG_ERROR, "OOM: %m");
+		goto ctx_alloc_failure;
+		return;
+	}
+
+	ctx->fd = fd;
+	ctx->length = len - ret;
+	ctx->index = 0;
+	ctx->data = malloc(ctx->length);
+	if (!ctx->data) {
+		nvnc_log(NVNC_LOG_ERROR, "OOM: %m");
+		goto ctx_data_alloc_failure;
+	}
+	memcpy(ctx->data, d + ret, ctx->length);
+
+	ctx->handler = aml_handler_new(ctx->fd, on_send, ctx, NULL);
+	if (!ctx->handler)
+		goto handler_failure;
+
+	aml_set_event_mask(ctx->handler, AML_EVENT_WRITE);
+
+	if (aml_start(aml_get_default(), ctx->handler) < 0)
+		goto poll_start_failure;
+
+	LIST_INSERT_HEAD(&self->send_contexts, ctx, link);
+	return;
+
+poll_start_failure:
+	aml_unref(ctx->handler);
+handler_failure:
+	free(ctx->data);
+ctx_data_alloc_failure:
+	free(ctx);
+ctx_alloc_failure:
 	close(fd);
+	nvnc_log(NVNC_LOG_ERROR, "Clipboard write incomplete");
 }
 
 static void data_control_source_cancelled(void* data,
@@ -235,6 +390,7 @@ static struct zwlr_data_control_source_v1* set_selection(struct data_control* se
 
 	zwlr_data_control_source_v1_add_listener(selection, &data_control_source_listener, self);
 	zwlr_data_control_source_v1_offer(selection, self->mime_type);
+	zwlr_data_control_source_v1_offer(selection, self->custom_mime_type_name);
 
 	if (primary)
 		zwlr_data_control_device_v1_set_primary_selection(self->device, selection);
@@ -244,21 +400,33 @@ static struct zwlr_data_control_source_v1* set_selection(struct data_control* se
 	return selection;
 }
 
-void data_control_init(struct data_control* self, struct wl_display* wl_display, struct nvnc* server, struct wl_seat* seat)
+void data_control_init(struct data_control* self, struct nvnc* server, struct wl_seat* seat)
 {
-	self->wl_display = wl_display;
 	self->server = server;
+	LIST_INIT(&self->receive_contexts);
+	LIST_INIT(&self->send_contexts);
 	self->device = zwlr_data_control_manager_v1_get_data_device(self->manager, seat);
 	zwlr_data_control_device_v1_add_listener(self->device, &data_control_device_listener, self);
 	self->selection = NULL;
 	self->primary_selection = NULL;
+	self->offer = NULL;
+	self->is_own_offer = false;
 	self->cb_data = NULL;
 	self->cb_len = 0;
 	self->mime_type = "text/plain;charset=utf-8";
+	snprintf(self->custom_mime_type_name,
+			sizeof(self->custom_mime_type_name),
+			"x-wayvnc-client-%08x", (unsigned int)rand());
 }
 
 void data_control_destroy(struct data_control* self)
 {
+	while (!LIST_EMPTY(&self->receive_contexts))
+		destroy_receive_context(LIST_FIRST(&self->receive_contexts));
+	while (!LIST_EMPTY(&self->send_contexts)) {
+		nvnc_log(NVNC_LOG_ERROR, "Clipboard write incomplete due to client disconnection");
+		destroy_send_context(LIST_FIRST(&self->send_contexts));
+	}
 	if (self->selection) {
 		zwlr_data_control_source_v1_destroy(self->selection);
 		self->selection = NULL;
