@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019 - 2023 Andri Yngvason
+ * Copyright (c) 2019 - 2024 Andri Yngvason
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -19,6 +19,7 @@
 #include <string.h>
 #include <stdint.h>
 #include <stdbool.h>
+#include <time.h>
 #include <unistd.h>
 #include <assert.h>
 #include <inttypes.h>
@@ -35,6 +36,8 @@
 #include <fcntl.h>
 
 #include "wlr-screencopy-unstable-v1.h"
+#include "ext-image-copy-capture-v1.h"
+#include "ext-image-capture-source-v1.h"
 #include "wlr-virtual-pointer-unstable-v1.h"
 #include "virtual-keyboard-unstable-v1.h"
 #include "xdg-output-unstable-v1.h"
@@ -42,7 +45,7 @@
 #include "wlr-output-management-unstable-v1.h"
 #include "linux-dmabuf-unstable-v1.h"
 #include "ext-transient-seat-v1.h"
-#include "screencopy.h"
+#include "screencopy-interface.h"
 #include "data-control.h"
 #include "strlcpy.h"
 #include "output.h"
@@ -56,16 +59,13 @@
 #include "ctl-server.h"
 #include "util.h"
 #include "option-parser.h"
+#include "pixels.h"
+#include "buffer.h"
 
 #ifdef ENABLE_PAM
 #include "pam_auth.h"
 #endif
 
-#ifdef ENABLE_SCREENCOPY_DMABUF
-#include <gbm.h>
-#include <xf86drm.h>
-#endif
-
 #define DEFAULT_ADDRESS "127.0.0.1"
 #define DEFAULT_PORT 5900
 
@@ -80,6 +80,7 @@ enum socket_type {
 	SOCKET_TYPE_TCP = 0,
 	SOCKET_TYPE_UNIX,
 	SOCKET_TYPE_WEBSOCKET,
+	SOCKET_TYPE_FROM_FD,
 };
 
 struct wayvnc {
@@ -100,7 +101,7 @@ struct wayvnc {
 	struct output* selected_output;
 	struct seat* selected_seat;
 
-	struct screencopy screencopy;
+	struct screencopy* screencopy;
 
 	struct aml_handler* wayland_handler;
 	struct aml_signal* signal_handler;
@@ -126,8 +127,15 @@ struct wayvnc {
 	bool is_initializing;
 
 	bool start_detached;
+	bool overlay_cursor;
+	int max_rate;
+	bool enable_gpu_features;
+	bool enable_resizing;
 
 	struct wayvnc_client* master_layout_client;
+
+	struct wayvnc_client* cursor_master;
+	struct screencopy* cursor_sc;
 };
 
 struct wayvnc_client {
@@ -144,7 +152,10 @@ struct wayvnc_client {
 };
 
 void wayvnc_exit(struct wayvnc* self);
-void on_capture_done(struct screencopy* sc);
+void on_capture_done(enum screencopy_result result, struct wv_buffer* buffer,
+		void* userdata);
+static void on_cursor_capture_done(enum screencopy_result result,
+		struct wv_buffer* buffer, void* userdata);
 static void on_nvnc_client_new(struct nvnc_client* client);
 void switch_to_output(struct wayvnc*, struct output*);
 void switch_to_next_output(struct wayvnc*);
@@ -158,12 +169,19 @@ static int blank_screen(struct wayvnc* self);
 static bool wayland_attach(struct wayvnc* self, const char* display,
 		const char* output);
 static void wayland_detach(struct wayvnc* self);
+static bool configure_cursor_sc(struct wayvnc* self,
+		struct wayvnc_client* client);
 
 struct wl_shm* wl_shm = NULL;
 struct zwp_linux_dmabuf_v1* zwp_linux_dmabuf = NULL;
-struct gbm_device* gbm_device = NULL;
 struct zxdg_output_manager_v1* xdg_output_manager = NULL;
 struct zwlr_output_power_manager_v1* wlr_output_power_manager = NULL;
+struct zwlr_screencopy_manager_v1* screencopy_manager = NULL;
+struct ext_output_image_capture_source_manager_v1*
+		ext_output_image_capture_source_manager = NULL;
+struct ext_image_copy_capture_manager_v1* ext_image_copy_capture_manager = NULL;
+
+extern struct screencopy_impl wlr_screencopy_impl, ext_image_copy_capture_impl;
 
 static bool registry_add_input(void* data, struct wl_registry* registry,
 			 uint32_t id, const char* interface,
@@ -240,7 +258,10 @@ static void registry_add(void* data, struct wl_registry* registry,
 		if (!self->is_initializing) {
 			wl_display_dispatch(self->display);
 			wl_display_roundtrip(self->display);
+
+			ctl_server_event_output_added(self->ctl, output->name);
 		}
+
 		return;
 	}
 
@@ -250,7 +271,7 @@ static void registry_add(void* data, struct wl_registry* registry,
 			wl_registry_bind(registry, id,
 			                 &zxdg_output_manager_v1_interface, 3);
 
-		output_setup_wl_managers(&self->outputs);
+		output_setup_xdg_output_managers(&self->outputs);
 		return;
 	}
 
@@ -258,7 +279,6 @@ static void registry_add(void* data, struct wl_registry* registry,
 		nvnc_trace("Registering new wlr_output_power_manager");
 		wlr_output_power_manager = wl_registry_bind(registry, id,
 					&zwlr_output_power_manager_v1_interface, 1);
-		output_setup_wl_managers(&self->outputs);
 		return;
 	}
 
@@ -272,13 +292,31 @@ static void registry_add(void* data, struct wl_registry* registry,
 	}
 
 	if (strcmp(interface, zwlr_screencopy_manager_v1_interface.name) == 0) {
-		self->screencopy.manager =
+		screencopy_manager =
 			wl_registry_bind(registry, id,
 					 &zwlr_screencopy_manager_v1_interface,
 					 MIN(3, version));
 		return;
 	}
 
+#if 1
+	if (strcmp(interface, ext_image_copy_capture_manager_v1_interface.name) == 0) {
+		ext_image_copy_capture_manager =
+			wl_registry_bind(registry, id,
+					 &ext_image_copy_capture_manager_v1_interface,
+					 MIN(1, version));
+		return;
+	}
+
+	if (strcmp(interface, ext_output_image_capture_source_manager_v1_interface.name) == 0) {
+		ext_output_image_capture_source_manager =
+			wl_registry_bind(registry, id,
+					&ext_output_image_capture_source_manager_v1_interface,
+					MIN(1, version));
+		return;
+	}
+#endif
+
 	if (strcmp(interface, wl_shm_interface.name) == 0) {
 		wl_shm = wl_registry_bind(registry, id, &wl_shm_interface, 1);
 		return;
@@ -322,6 +360,8 @@ static void registry_remove(void* data, struct wl_registry* registry,
 		} else
 			nvnc_log(NVNC_LOG_INFO, "Output %s went away", out->name);
 
+		ctl_server_event_output_removed(self->ctl, out->name);
+
 		wl_list_remove(&out->link);
 		output_destroy(out);
 
@@ -348,46 +388,6 @@ static void registry_remove(void* data, struct wl_registry* registry,
 	}
 }
 
-#ifdef ENABLE_SCREENCOPY_DMABUF
-static int find_render_node(char *node, size_t maxlen) {
-	bool r = -1;
-	drmDevice *devices[64];
-
-	int n = drmGetDevices2(0, devices, sizeof(devices) / sizeof(devices[0]));
-	for (int i = 0; i < n; ++i) {
-		drmDevice *dev = devices[i];
-		if (!(dev->available_nodes & (1 << DRM_NODE_RENDER)))
-			continue;
-
-		strlcpy(node, dev->nodes[DRM_NODE_RENDER], maxlen);
-		r = 0;
-		break;
-	}
-
-	drmFreeDevices(devices, n);
-	return r;
-}
-
-static int init_render_node(int* fd)
-{
-	char render_node[256];
-	if (find_render_node(render_node, sizeof(render_node)) < 0)
-		return -1;
-
-	*fd = open(render_node, O_RDWR);
-	if (*fd < 0)
-		return -1;
-
-	gbm_device = gbm_create_device(*fd);
-	if (!gbm_device) {
-		close(*fd);
-		return -1;
-	}
-
-	return 0;
-}
-#endif
-
 static void wayland_detach(struct wayvnc* self)
 {
 	if (!self->display)
@@ -412,7 +412,6 @@ static void wayland_detach(struct wayvnc* self)
 	}
 
 	self->selected_output = NULL;
-	self->screencopy.wl_output = NULL;
 
 	output_list_destroy(&self->outputs);
 	seat_list_destroy(&self->seats);
@@ -421,12 +420,13 @@ static void wayland_detach(struct wayvnc* self)
 		zwp_linux_dmabuf_v1_destroy(zwp_linux_dmabuf);
 	zwp_linux_dmabuf = NULL;
 
-	if (self->screencopy.manager) {
-		screencopy_stop(&self->screencopy);
-		screencopy_destroy(&self->screencopy);
-		zwlr_screencopy_manager_v1_destroy(self->screencopy.manager);
-	}
-	self->screencopy.manager = NULL;
+	screencopy_stop(self->screencopy);
+	screencopy_destroy(self->screencopy);
+	self->screencopy = NULL;
+
+	screencopy_stop(self->cursor_sc);
+	screencopy_destroy(self->cursor_sc);
+	self->cursor_sc = NULL;
 
 	if (xdg_output_manager)
 		zxdg_output_manager_v1_destroy(xdg_output_manager);
@@ -459,10 +459,28 @@ static void wayland_detach(struct wayvnc* self)
 	}
 	self->performance_ticker = NULL;
 
-	if (self->capture_retry_timer)
+	if (screencopy_manager)
+		zwlr_screencopy_manager_v1_destroy(screencopy_manager);
+	screencopy_manager = NULL;
+
+	if (ext_output_image_capture_source_manager)
+		ext_output_image_capture_source_manager_v1_destroy(
+				ext_output_image_capture_source_manager);
+	ext_output_image_capture_source_manager = NULL;
+
+	if (ext_image_copy_capture_manager)
+		ext_image_copy_capture_manager_v1_destroy(ext_image_copy_capture_manager);
+	ext_image_copy_capture_manager = NULL;
+
+	if (self->capture_retry_timer) {
+		aml_stop(aml_get_default(), self->capture_retry_timer);
 		aml_unref(self->capture_retry_timer);
+	}
 	self->capture_retry_timer = NULL;
 
+	if (self->transient_seat_manager)
+		ext_transient_seat_manager_v1_destroy(self->transient_seat_manager);
+
 	wl_registry_destroy(self->registry);
 	self->registry = NULL;
 
@@ -554,7 +572,7 @@ static int init_wayland(struct wayvnc* self, const char* display)
 		goto failure;
 	}
 
-	if (!self->screencopy.manager) {
+	if (!screencopy_manager && !ext_image_copy_capture_manager) {
 		nvnc_log(NVNC_LOG_ERROR, "Screencopy protocol not supported by compositor. Exiting. Refer to FAQ section in man page.");
 		goto failure;
 	}
@@ -564,9 +582,6 @@ static int init_wayland(struct wayvnc* self, const char* display)
 		goto failure;
 	}
 
-	self->screencopy.on_done = on_capture_done;
-	self->screencopy.userdata = self;
-
 	self->wl_handler = aml_handler_new(wl_display_get_fd(self->display),
 	                             on_wayland_event, self, NULL);
 	if (!self->wl_handler)
@@ -640,9 +655,9 @@ static void compose_client_info(const struct wayvnc_client* client,
 		struct ctl_server_client_info* info)
 {
 	info->id = client->id;
-	socklen_t addrlen = sizeof(info->address);
+	socklen_t addrlen = sizeof(info->address_storage);
 	nvnc_client_get_address(client->nvnc_client,
-			(struct sockaddr*)&info->address, &addrlen);
+			(struct sockaddr*)&info->address_storage, &addrlen);
 	info->username = nvnc_client_get_auth_username(client->nvnc_client);
 	info->seat = client->seat ? client->seat->name : NULL;
 }
@@ -755,6 +770,9 @@ static void on_key_event(struct nvnc_client* client, uint32_t symbol,
 	}
 
 	keyboard_feed(&wv_client->keyboard, symbol, is_pressed);
+
+	nvnc_client_set_led_state(wv_client->nvnc_client,
+			keyboard_get_led_state(&wv_client->keyboard));
 }
 
 static void on_key_code_event(struct nvnc_client* client, uint32_t code,
@@ -766,6 +784,9 @@ static void on_key_code_event(struct nvnc_client* client, uint32_t code,
 	}
 
 	keyboard_feed_code(&wv_client->keyboard, code + 8, is_pressed);
+
+	nvnc_client_set_led_state(wv_client->nvnc_client,
+			keyboard_get_led_state(&wv_client->keyboard));
 }
 
 static void on_client_cut_text(struct nvnc_client* nvnc_client,
@@ -886,21 +907,29 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 		case SOCKET_TYPE_WEBSOCKET:
 			self->nvnc = nvnc_open_websocket(addr, port);
 			break;
+		case SOCKET_TYPE_FROM_FD:;
+			int fd = atoi(addr);
+			self->nvnc = nvnc_open_from_fd(fd);
+			break;
 		default:
 			abort();
 	}
 	if (!self->nvnc) {
-		nvnc_log(NVNC_LOG_ERROR, "Failed to bind to address. Add -Ldebug to the argument list for more info.");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to listen on socket or bind to its address. Add -Ldebug to the argument list for more info.");
 		return -1;
 	}
 	if (socket_type == SOCKET_TYPE_UNIX)
 		nvnc_log(NVNC_LOG_INFO, "Listening for connections on %s", addr);
+	else if (socket_type == SOCKET_TYPE_FROM_FD)
+		nvnc_log(NVNC_LOG_INFO, "Listening for connections on fd %s",
+				addr);
 	else
-		nvnc_log(NVNC_LOG_INFO, "Listening for connections on %s:%d", addr, port);
+		nvnc_log(NVNC_LOG_INFO, "Listening for connections on %s:%d",
+				addr, port);
 
 	self->nvnc_display = nvnc_display_new(0, 0);
 	if (!self->nvnc_display)
-		goto failure;
+		goto display_failure;
 
 	nvnc_add_display(self->nvnc, self->nvnc_display);
 
@@ -908,7 +937,8 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 
 	nvnc_set_name(self->nvnc, "WayVNC");
 
-	nvnc_set_desktop_layout_fn(self->nvnc, on_client_resize);
+	if (self->enable_resizing)
+		nvnc_set_desktop_layout_fn(self->nvnc, on_client_resize);
 
 	enum nvnc_auth_flags auth_flags = 0;
 	if (self->cfg.enable_auth) {
@@ -921,7 +951,7 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 	if (self->cfg.enable_auth) {
 		if (nvnc_enable_auth(self->nvnc, auth_flags, on_auth, self) < 0) {
 			nvnc_log(NVNC_LOG_ERROR, "Failed to enable authentication");
-			goto failure;
+			goto auth_failure;
 		}
 
 		if (self->cfg.rsa_private_key_file) {
@@ -930,7 +960,7 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 					self->cfg.rsa_private_key_file);
 			if (nvnc_set_rsa_creds(self->nvnc, key_file) < 0) {
 				nvnc_log(NVNC_LOG_ERROR, "Failed to load RSA credentials");
-				goto failure;
+				goto auth_failure;
 			}
 		}
 
@@ -946,7 +976,7 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 					cert_file);
 			if (r < 0) {
 				nvnc_log(NVNC_LOG_ERROR, "Failed to enable TLS authentication");
-				goto failure;
+				goto auth_failure;
 			}
 		}
 	}
@@ -960,18 +990,28 @@ static int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port,
 	nvnc_set_cut_text_fn(self->nvnc, on_client_cut_text);
 
 	if (blank_screen(self) != 0)
-		goto failure;
+		goto blank_screen_failure;
 
 	return 0;
 
-failure:
+blank_screen_failure:
+auth_failure:
+	nvnc_display_unref(self->nvnc_display);
+display_failure:
 	nvnc_close(self->nvnc);
 	return -1;
 }
 
+static void wayvnc_start_cursor_capture(struct wayvnc* self, bool immediate)
+{
+	if (self->cursor_sc) {
+		screencopy_start(self->cursor_sc, immediate);
+	}
+}
+
 int wayvnc_start_capture(struct wayvnc* self)
 {
-	int rc = screencopy_start(&self->screencopy);
+	int rc = screencopy_start(self->screencopy, false);
 	if (rc < 0) {
 		nvnc_log(NVNC_LOG_ERROR, "Failed to start capture. Exiting...");
 		wayvnc_exit(self);
@@ -984,15 +1024,19 @@ int wayvnc_start_capture_immediate(struct wayvnc* self)
 	if (self->capture_retry_timer)
 		return 0;
 
-	if (self->selected_output->power == OUTPUT_POWER_OFF) {
-		nvnc_log(NVNC_LOG_WARNING, "Selected output is in powersaving mode. Delaying capture until it turns on.");
-		if (output_set_power_state(self->selected_output, OUTPUT_POWER_ON)
-				== 0)
-			nvnc_log(NVNC_LOG_WARNING, "Requested power ON.");
+	struct output* output = self->selected_output;
+	int rc = output_acquire_power_on(output);
+	if (rc == 0) {
+		nvnc_log(NVNC_LOG_DEBUG, "Acquired power state management. Waiting for power event to start capturing");
 		return 0;
+	} else if (rc > 0 && output->power != OUTPUT_POWER_ON) {
+		nvnc_log(NVNC_LOG_DEBUG, "Output power state management already acquired, but not yet powered on");
+		return 0;
+	} else if (rc < 0) {
+		nvnc_log(NVNC_LOG_WARNING, "Failed to acquire power state control. Capturing may fail.");
 	}
 
-	int rc = screencopy_start_immediate(&self->screencopy);
+	rc = screencopy_start(self->screencopy, true);
 	if (rc < 0) {
 		nvnc_log(NVNC_LOG_ERROR, "Failed to start capture. Exiting...");
 		wayvnc_exit(self);
@@ -1030,13 +1074,14 @@ void on_output_dimension_change(struct output* output)
 
 	nvnc_log(NVNC_LOG_DEBUG, "Output dimensions changed. Restarting frame capturer...");
 
-	screencopy_stop(&self->screencopy);
+	screencopy_stop(self->screencopy);
 	wayvnc_start_capture_immediate(self);
 }
 
 static void on_output_power_change(struct output* output)
 {
-	nvnc_trace("Output %s power state changed to %s", output->name, output_power_state_name(output->power));
+	nvnc_trace("Output %s power state changed to %s", output->name,
+			output_power_state_name(output->power));
 
 	struct wayvnc* self = output->userdata;
 	if (self->selected_output != output || self->nr_clients == 0)
@@ -1044,12 +1089,13 @@ static void on_output_power_change(struct output* output)
 
 	switch (output->power) {
 	case OUTPUT_POWER_ON:
-		nvnc_log(NVNC_LOG_WARNING, "Output is now on. Restarting frame capture");
 		wayvnc_start_capture_immediate(self);
+		wayvnc_start_cursor_capture(self, true);
 		break;
 	case OUTPUT_POWER_OFF:
 		nvnc_log(NVNC_LOG_WARNING, "Output is now off. Pausing frame capture");
-		screencopy_stop(&self->screencopy);
+		screencopy_stop(self->cursor_sc);
+		screencopy_stop(self->screencopy);
 		blank_screen(self);
 		break;
 	default:
@@ -1057,35 +1103,9 @@ static void on_output_power_change(struct output* output)
 	}
 }
 
-static uint32_t calculate_region_area(struct pixman_region16* region)
+static void apply_output_transform(const struct wayvnc* self,
+		struct wv_buffer* buffer, struct pixman_region16* damage)
 {
-	uint32_t area = 0;
-
-	int n_rects = 0;
-	struct pixman_box16* rects = pixman_region_rectangles(region,
-		&n_rects);
-
-	for (int i = 0; i < n_rects; ++i) {
-		int width = rects[i].x2 - rects[i].x1;
-		int height = rects[i].y2 - rects[i].y1;
-		area += width * height;
-	}
-
-	return area;
-}
-
-void wayvnc_process_frame(struct wayvnc* self)
-{
-	struct wv_buffer* buffer = self->screencopy.back;
-	self->screencopy.back = NULL;
-
-	self->n_frames_captured++;
-	self->damage_area_sum +=
-		calculate_region_area(&buffer->damage);
-
-	struct pixman_region16 damage;
-	pixman_region_init(&damage);
-
 	enum wl_output_transform output_transform, buffer_transform;
 	output_transform = self->selected_output->transform;
 
@@ -1093,16 +1113,34 @@ void wayvnc_process_frame(struct wayvnc* self)
 		buffer_transform = wv_output_transform_compose(output_transform,
 				WL_OUTPUT_TRANSFORM_FLIPPED_180);
 
-		wv_region_transform(&damage, &buffer->damage,
+		wv_region_transform(damage, &buffer->frame_damage,
 				WL_OUTPUT_TRANSFORM_FLIPPED_180,
 				buffer->width, buffer->height);
 	} else {
 		buffer_transform = output_transform;
-		pixman_region_copy(&damage, &buffer->damage);
+		pixman_region_copy(damage, &buffer->frame_damage);
 	}
 
 	nvnc_fb_set_transform(buffer->nvnc_fb,
 			(enum nvnc_transform)buffer_transform);
+}
+
+void wayvnc_process_frame(struct wayvnc* self, struct wv_buffer* buffer)
+{
+	nvnc_trace("Passing on buffer: %p", buffer);
+
+	self->n_frames_captured++;
+	self->damage_area_sum +=
+		calculate_region_area(&buffer->frame_damage);
+
+	struct pixman_region16 damage;
+	pixman_region_init(&damage);
+
+	if (self->screencopy->impl->caps & SCREENCOPY_CAP_TRANSFORM) {
+		pixman_region_copy(&damage, &buffer->frame_damage);
+	} else {
+		apply_output_transform(self, buffer, &damage);
+	}
 
 	pixman_region_intersect_rect(&damage, &damage, 0, 0, buffer->width,
 			buffer->height);
@@ -1115,15 +1153,12 @@ void wayvnc_process_frame(struct wayvnc* self)
 	wayvnc_start_capture(self);
 }
 
-void on_capture_done(struct screencopy* sc)
+void on_capture_done(enum screencopy_result result, struct wv_buffer* buffer,
+		void* userdata)
 {
-	struct wayvnc* self = sc->userdata;
+	struct wayvnc* self = userdata;
 
-	switch (sc->status) {
-	case SCREENCOPY_STOPPED:
-		break;
-	case SCREENCOPY_IN_PROGRESS:
-		break;
+	switch (result) {
 	case SCREENCOPY_FATAL:
 		nvnc_log(NVNC_LOG_ERROR, "Fatal error while capturing. Exiting...");
 		wayvnc_exit(self);
@@ -1132,7 +1167,7 @@ void on_capture_done(struct screencopy* sc)
 		wayvnc_restart_capture(self);
 		break;
 	case SCREENCOPY_DONE:
-		wayvnc_process_frame(self);
+		wayvnc_process_frame(self, buffer);
 		break;
 	}
 }
@@ -1259,6 +1294,9 @@ static struct wayvnc_client* client_create(struct wayvnc* wayvnc,
 
 	self->id = next_client_id++;
 
+	if (!wayvnc->cursor_master)
+		wayvnc->cursor_master = self;
+
 	if (wayvnc->display) {
 		client_init_wayland(self);
 	}
@@ -1275,6 +1313,14 @@ static void client_destroy(void* obj)
 	if (self == wayvnc->master_layout_client)
 		wayvnc->master_layout_client = NULL;
 
+	if (self == wayvnc->cursor_master) {
+		nvnc_set_cursor(wayvnc->nvnc, NULL, 0, 0, 0, 0, false);
+		screencopy_stop(wayvnc->cursor_sc);
+		screencopy_destroy(wayvnc->cursor_sc);
+		wayvnc->cursor_sc = NULL;
+		wayvnc->cursor_master = NULL;
+	}
+
 	if (self->transient_seat)
 		ext_transient_seat_v1_destroy(self->transient_seat);
 
@@ -1295,7 +1341,8 @@ static void client_destroy(void* obj)
 
 	if (wayvnc->nr_clients == 0 && wayvnc->display) {
 		nvnc_log(NVNC_LOG_INFO, "Stopping screen capture");
-		screencopy_stop(&wayvnc->screencopy);
+		screencopy_stop(wayvnc->screencopy);
+		output_release_power_on(wayvnc->selected_output);
 		stop_performance_ticker(wayvnc);
 	}
 
@@ -1330,10 +1377,9 @@ static void on_nvnc_client_new(struct nvnc_client* client)
 	assert(wayvnc_client);
 	nvnc_set_userdata(client, wayvnc_client, client_destroy);
 
-	if (self->nr_clients == 0 && self->display) {
+	if (self->nr_clients++ == 0 && self->display) {
 		handle_first_client(self);
 	}
-	self->nr_clients++;
 	nvnc_log(NVNC_LOG_DEBUG, "Client connected, new client count: %d",
 			self->nr_clients);
 
@@ -1381,6 +1427,17 @@ static void client_init_pointer(struct wayvnc_client* self)
 	if (pointer_init(&self->pointer) < 0) {
 		nvnc_log(NVNC_LOG_ERROR, "Failed to initialise pointer");
 	}
+
+	if (self == wayvnc->cursor_master) {
+		// Get seat capability update
+		// TODO: Make this asynchronous
+		wl_display_roundtrip(wayvnc->display);
+		wl_display_dispatch_pending(wayvnc->display);
+
+		configure_cursor_sc(wayvnc, self);
+		if (wayvnc->cursor_sc)
+			screencopy_start(wayvnc->cursor_sc, true);
+	}
 }
 
 static void handle_transient_seat_ready(void* data,
@@ -1494,7 +1551,7 @@ static void client_init_data_control(struct wayvnc_client* self)
 		return;
 
 	self->data_control.manager = wayvnc->data_control_manager;
-	data_control_init(&self->data_control, wayvnc->display, wayvnc->nvnc,
+	data_control_init(&self->data_control, wayvnc->nvnc,
 			self->seat->wl_seat);
 }
 
@@ -1514,15 +1571,135 @@ void log_selected_output(struct wayvnc* self)
 	}
 }
 
-void set_selected_output(struct wayvnc* self, struct output* output) {
+static void wayvnc_process_cursor(struct wayvnc* self, struct wv_buffer* buffer)
+{
+	nvnc_log(NVNC_LOG_DEBUG, "Got new cursor");
+	bool is_damaged = pixman_region_not_empty(&buffer->frame_damage);
+	nvnc_set_cursor(self->nvnc, buffer->nvnc_fb, buffer->width,
+			buffer->height, buffer->x_hotspot, buffer->y_hotspot,
+			is_damaged);
+	wayvnc_start_cursor_capture(self, false);
+}
+
+static void on_cursor_capture_done(enum screencopy_result result,
+		struct wv_buffer* buffer, void* userdata)
+{
+	struct wayvnc* self = userdata;
+
+	switch (result) {
+	case SCREENCOPY_FATAL:
+		nvnc_log(NVNC_LOG_ERROR, "Fatal error while capturing. Exiting...");
+		wayvnc_exit(self);
+		break;
+	case SCREENCOPY_FAILED:
+		wayvnc_start_cursor_capture(self, true);
+		break;
+	case SCREENCOPY_DONE:
+		wayvnc_process_cursor(self, buffer);
+		break;
+	}
+}
+
+static double rate_format(const void* userdata, enum wv_buffer_type type,
+		enum wv_buffer_domain domain, uint32_t format,
+		uint64_t modifier)
+{
+	const struct wayvnc* self = userdata;
+
+	enum nvnc_fb_type fb_type = NVNC_FB_UNSPEC;
+	switch (type) {
+	case WV_BUFFER_SHM:
+		fb_type = NVNC_FB_SIMPLE;
+		break;
+#ifdef ENABLE_SCREENCOPY_DMABUF
+	case WV_BUFFER_DMABUF:
+		fb_type = NVNC_FB_GBM_BO;
+		break;
+#endif
+	case WV_BUFFER_UNSPEC:;
+	}
+	assert(fb_type != NVNC_FB_UNSPEC);
+
+	switch (domain) {
+	case WV_BUFFER_DOMAIN_OUTPUT:
+		return nvnc_rate_pixel_format(self->nvnc, fb_type, format,
+				modifier);
+	case WV_BUFFER_DOMAIN_CURSOR:
+		return nvnc_rate_cursor_pixel_format(self->nvnc, fb_type, format,
+				modifier);
+	case WV_BUFFER_DOMAIN_UNSPEC:;
+	}
+
+	abort();
+	return -1;
+}
+
+static bool configure_cursor_sc(struct wayvnc* self,
+		struct wayvnc_client* client)
+{
+	nvnc_log(NVNC_LOG_DEBUG, "Configuring cursor capturing");
+
+	screencopy_stop(self->cursor_sc);
+	screencopy_destroy(self->cursor_sc);
+
+	struct seat* seat = client->seat;
+	assert(seat);
+
+	if (!(seat->capabilities & WL_SEAT_CAPABILITY_POINTER)) {
+		nvnc_log(NVNC_LOG_DEBUG, "Client's seat has no pointer capability");
+		return false;
+	}
+
+	self->cursor_sc = screencopy_create_cursor(
+			self->selected_output->wl_output, seat->wl_seat);
+	if (!self->cursor_sc) {
+		nvnc_log(NVNC_LOG_DEBUG, "Failed to capture cursor");
+		return false;
+	}
+
+	self->cursor_sc->on_done = on_cursor_capture_done;
+	self->cursor_sc->rate_format = rate_format;
+	self->cursor_sc->userdata = self;
+
+	self->cursor_sc->rate_limit = self->max_rate;
+	self->cursor_sc->enable_linux_dmabuf = false;
+
+	nvnc_log(NVNC_LOG_DEBUG, "Configured cursor capturing");
+	return true;
+}
+
+bool configure_screencopy(struct wayvnc* self)
+{
+	screencopy_stop(self->screencopy);
+	screencopy_destroy(self->screencopy);
+
+	self->screencopy = screencopy_create(self->selected_output->wl_output,
+			self->overlay_cursor);
+	if (!self->screencopy) {
+		nvnc_log(NVNC_LOG_ERROR, "screencopy is not supported by compositor");
+		return false;
+	}
+
+	self->screencopy->on_done = on_capture_done;
+	self->screencopy->rate_format = rate_format;
+	self->screencopy->userdata = self;
+
+	self->screencopy->rate_limit = self->max_rate;
+	self->screencopy->enable_linux_dmabuf = self->enable_gpu_features;
+
+	return true;
+}
+
+void set_selected_output(struct wayvnc* self, struct output* output)
+{
 	if (self->selected_output) {
 		self->selected_output->on_dimension_change = NULL;
 	}
 	self->selected_output = output;
-	self->screencopy.wl_output = output->wl_output;
 	output->on_dimension_change = on_output_dimension_change;
 	output->on_power_change = on_output_power_change;
 	output->userdata = self;
+
 	if (self->ctl)
 		ctl_server_event_capture_changed(self->ctl, output->name);
 	log_selected_output(self);
@@ -1535,11 +1712,16 @@ void switch_to_output(struct wayvnc* self, struct output* output)
 				output->name);
 		return;
 	}
-	screencopy_stop(&self->screencopy);
+	screencopy_stop(self->screencopy);
+	output_release_power_on(output);
 	set_selected_output(self, output);
+	configure_screencopy(self);
 	reinitialise_pointers(self);
 	if (self->nr_clients > 0)
 		wayvnc_start_capture_immediate(self);
+	screencopy_stop(self->cursor_sc);
+	if (self->cursor_sc)
+		screencopy_start(self->cursor_sc, true);
 }
 
 void switch_to_next_output(struct wayvnc* self)
@@ -1627,14 +1809,13 @@ static bool wayland_attach(struct wayvnc* self, const char* display,
 		}
 	}
 
-	screencopy_init(&self->screencopy);
-	if (!self->screencopy.manager) {
+	if (!screencopy_manager) {
 		nvnc_log(NVNC_LOG_ERROR, "Attached display does not implement wlr-screencopy-v1");
 		wayland_detach(self);
 		return false;
 	}
-
 	set_selected_output(self, out);
+	configure_screencopy(self);
 
 	struct nvnc_client* nvnc_client;
 	for (nvnc_client = nvnc_client_first(self->nvnc); nvnc_client;
@@ -1694,6 +1875,7 @@ int main(int argc, char* argv[])
 
 	const char* address = NULL;
 	int port = 0;
+	bool use_external_fd = false;
 	bool use_unix_socket = false;
 	bool use_websocket = false;
 	bool start_detached = false;
@@ -1709,8 +1891,6 @@ int main(int argc, char* argv[])
 	bool disable_input = false;
 	bool use_transient_seat = false;
 
-	int drm_fd MAYBE_UNUSED = -1;
-
 	int log_level = NVNC_LOG_WARNING;
 
 	static const struct wv_option opts[] = {
@@ -1722,42 +1902,46 @@ int main(int argc, char* argv[])
 		  .default_ = XSTR(DEFAULT_PORT)},
 		{ 'C', "config", "<path>",
 		  "Select a config file." },
+		{ 'd', "disable-input", NULL,
+		  "Disable all remote input." },
+		{ 'D', "detached", NULL,
+		  "Start detached from a compositor." },
+		{ 'f', "max-fps", "<fps>",
+		  "Set rate limit.",
+		  .default_ = "30" },
 		{ 'g', "gpu", NULL,
 		  "Enable features that need GPU." },
-		{ 'o', "output", "<name>",
-		  "Select output to capture." },
+		{ 'h', "help", NULL,
+		  "Get help (this text)." },
 		{ 'k', "keyboard", "<layout>[-<variant>]",
 		  "Select keyboard layout with an optional variant." },
+		{ 'L', "log-level", "<level>",
+		  "Set log level. The levels are: error, warning, info, debug trace and quiet.",
+		  .default_ = "warning" },
+		{ 'o', "output", "<name>",
+		  "Select output to capture." },
+		{ 'p', "show-performance", NULL,
+		  "Show performance counters." },
+		{ 'r', "render-cursor", NULL,
+		  "Enable overlay cursor rendering." },
+		{ 'R', "disable-resizing", NULL,
+		  "Disable automatic resizing." },
 		{ 's', "seat", "<name>",
 		  "Select seat by name." },
 		{ 'S', "socket", "<path>",
 		  "Control socket path." },
 		{ 't', "transient-seat", NULL,
 		  "Use transient seat." },
-		{ 'r', "render-cursor", NULL,
-		  "Enable overlay cursor rendering." },
-		{ 'f', "max-fps", "<fps>",
-		  "Set rate limit.",
-		  .default_ = "30" },
-		{ 'p', "performance", NULL,
-		  "Show performance counters." },
 		{ 'u', "unix-socket", NULL,
 		  "Create unix domain socket." },
-		{ 'd', "disable-input", NULL,
-		  "Disable all remote input." },
-		{ 'D', "detached", NULL,
-		  "Start detached from a compositor." },
-		{ 'V', "version", NULL,
-		  "Show version info." },
 		{ 'v', "verbose", NULL,
 		  "Be more verbose. Same as setting --log-level=info" },
+		{ 'V', "version", NULL,
+		  "Show version info." },
 		{ 'w', "websocket", NULL,
 		  "Create a websocket." },
-		{ 'L', "log-level", "<level>",
-		  "Set log level. The levels are: error, warning, info, debug trace and quiet.",
-		  .default_ = "warning" },
-		{ 'h', "help", NULL,
-		  "Get help (this text)." },
+		{ 'x', "external-listener-fd", NULL,
+		  "The address is a pre-bound file descriptor.", },
 		{}
 	};
 
@@ -1782,9 +1966,12 @@ int main(int argc, char* argv[])
 	seat_name = option_parser_get_value(&option_parser, "seat");
 	socket_path = option_parser_get_value(&option_parser, "socket");
 	overlay_cursor = !!option_parser_get_value(&option_parser, "render-cursor");
-	show_performance = !!option_parser_get_value(&option_parser, "performance");
+	show_performance = !!option_parser_get_value(&option_parser,
+			"show-performance");
 	use_unix_socket = !!option_parser_get_value(&option_parser, "unix-socket");
 	use_websocket = !!option_parser_get_value(&option_parser, "websocket");
+	use_external_fd = !!option_parser_get_value(&option_parser,
+			"external-listener-fd");
 	disable_input = !!option_parser_get_value(&option_parser, "disable-input");
 	log_level = option_parser_get_value(&option_parser, "verbose")
 		? NVNC_LOG_INFO : NVNC_LOG_WARNING;
@@ -1794,8 +1981,13 @@ int main(int argc, char* argv[])
 	use_transient_seat = !!option_parser_get_value(&option_parser,
 				"transient-seat");
 	start_detached = !!option_parser_get_value(&option_parser, "detached");
+	self.enable_resizing = !option_parser_get_value(&option_parser,
+			"disable-resizing");
 
 	self.start_detached = start_detached;
+	self.overlay_cursor = overlay_cursor;
+	self.max_rate = max_rate;
+	self.enable_gpu_features = enable_gpu_features;
 
 	keyboard_options = option_parser_get_value(&option_parser, "keyboard");
 	if (keyboard_options)
@@ -1815,8 +2007,10 @@ int main(int argc, char* argv[])
 		return 1;
 	}
 
-	if (use_unix_socket && use_websocket) {
-		nvnc_log(NVNC_LOG_ERROR, "websocket and unix-socket are conflicting options");
+	int n_address_modifiers = use_unix_socket + use_websocket +
+		use_external_fd;
+	if (n_address_modifiers > 1) {
+		nvnc_log(NVNC_LOG_ERROR, "Only one of the websocket, unix-socket or the external-listener-fd options may be set");
 		return 1;
 	}
 
@@ -1826,7 +2020,7 @@ int main(int argc, char* argv[])
 	}
 
 	if (seat_name && use_transient_seat) {
-		nvnc_log(NVNC_LOG_ERROR, "transient seat and seat are conflicting options");
+		nvnc_log(NVNC_LOG_ERROR, "transient-seat and seat are conflicting options");
 		return 1;
 	}
 
@@ -1857,6 +2051,10 @@ int main(int argc, char* argv[])
 	self.disable_input = disable_input;
 	self.use_transient_seat = use_transient_seat;
 
+	srand(time(NULL));
+
+	signal(SIGPIPE, SIG_IGN);
+
 	struct aml* aml = aml_new();
 	if (!aml)
 		goto failure;
@@ -1900,15 +2098,6 @@ int main(int argc, char* argv[])
 		self.selected_seat = seat;
 	}
 
-	self.screencopy.rate_limit = max_rate;
-	self.screencopy.enable_linux_dmabuf = enable_gpu_features;
-
-#ifdef ENABLE_SCREENCOPY_DMABUF
-	if (enable_gpu_features && init_render_node(&drm_fd) < 0) {
-		nvnc_log(NVNC_LOG_ERROR, "Failed to initialise DRM render node. No GPU acceleration will be available.");
-	}
-#endif
-
 	if (aml_unstable_abi_version != AML_UNSTABLE_API)
 		nvnc_log(NVNC_LOG_PANIC, "libaml is incompatible with this build of wayvnc!");
 
@@ -1917,21 +2106,11 @@ int main(int argc, char* argv[])
 		socket_type = SOCKET_TYPE_UNIX;
 	else if (use_websocket)
 		socket_type = SOCKET_TYPE_WEBSOCKET;
+	else if (use_external_fd)
+		socket_type = SOCKET_TYPE_FROM_FD;
 
-	if (init_nvnc(&self, address, port, socket_type) < 0)
-		goto nvnc_failure;
-
-	if (!start_detached) {
-		if (self.screencopy.manager)
-			screencopy_init(&self.screencopy);
-
-		if (!self.screencopy.manager) {
-			nvnc_log(NVNC_LOG_ERROR, "screencopy is not supported by compositor");
-			goto capture_failure;
-		}
-	}
-
-	self.screencopy.overlay_cursor = overlay_cursor;
+	if (!start_detached && !configure_screencopy(&self))
+		goto screencopy_failure;
 
 	if (show_performance)
 		self.performance_ticker = aml_ticker_new(1000000, on_perf_tick,
@@ -1953,6 +2132,9 @@ int main(int argc, char* argv[])
 	if (!self.ctl)
 		goto ctl_server_failure;
 
+	if (init_nvnc(&self, address, port, socket_type) < 0)
+		goto nvnc_failure;
+
 	if (self.display)
 		wl_display_dispatch_pending(self.display);
 
@@ -1967,7 +2149,7 @@ int main(int argc, char* argv[])
 	nvnc_log(NVNC_LOG_INFO, "Exiting...");
 
 	if (self.display)
-		screencopy_stop(&self.screencopy);
+		screencopy_stop(self.screencopy);
 
 	ctl_server_destroy(self.ctl);
 	self.ctl = NULL;
@@ -1976,31 +2158,22 @@ int main(int argc, char* argv[])
 	nvnc_close(self.nvnc);
 	self.nvnc = NULL;
 	wayvnc_destroy(&self);
-#ifdef ENABLE_SCREENCOPY_DMABUF
-	if (gbm_device) {
-		gbm_device_destroy(gbm_device);
-		close(drm_fd);
-	}
-#endif
+	if (zwp_linux_dmabuf)
+		zwp_linux_dmabuf_v1_destroy(zwp_linux_dmabuf);
+	if (self.screencopy)
+		screencopy_destroy(self.screencopy);
 	aml_unref(aml);
 
 	return 0;
 
-ctl_server_failure:
-capture_failure:
-	nvnc_display_unref(self.nvnc_display);
-	nvnc_close(self.nvnc);
 nvnc_failure:
+	ctl_server_destroy(self.ctl);
+ctl_server_failure:
+screencopy_failure:
 wayland_failure:
 	aml_unref(aml);
 failure:
 	self.nvnc = NULL;
 	wayvnc_destroy(&self);
-#ifdef ENABLE_SCREENCOPY_DMABUF
-	if (gbm_device)
-		gbm_device_destroy(gbm_device);
-	if (drm_fd >= 0)
-		close(drm_fd);
-#endif
 	return 1;
 }
